# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import pytest
import hamcrest
import zope.interface.verify
import koppeltaal.definitions
import koppeltaal.interfaces
import koppeltaal.testing


def test_connector(connector):
    assert koppeltaal.interfaces.IConnector.providedBy(connector)
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IConnector, connector)


def test_search_message_id_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?_id=45909',
        respond_with='fixtures/bundle_one_message.json')

    models = list(connector.search(message_id='45909'))
    assert len(models) > 1

    message = models[0]
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.MessageHeader, message)
    assert message.event == 'CreateOrUpdateCarePlan'
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.CarePlan, message.data[0])
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Patient, message.patient)


def test_send_careplan_success_from_fixture(
        connector, transport, careplan_from_fixture):
    transport.expect(
        'POST',
        '/FHIR/Koppeltaal/Mailbox',
        respond_with='fixtures/bundle_post_careplan_ok.json')
    response_data = connector.send(
        'CreateOrUpdateCarePlan',
        careplan_from_fixture,
        careplan_from_fixture.patient)
    assert len(response_data) == 3
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource, response_data[0])
    assert response_data[0].fhir_link == (
        'https://example.com/fhir/Koppeltaal/CarePlan/1/'
        '_history/1970-01-01T01:01:01:01.1')
    assert response_data[1].fhir_link == (
        'https://example.com/fhir/Koppeltaal/Practitioner/1/'
        '_history/1970-01-01T01:01:01:01.1')
    assert response_data[2].fhir_link == (
        'https://example.com/fhir/Koppeltaal/Patient/1/'
        '_history/1970-01-01T01:01:01:01.1')

    response = transport.called.get('/FHIR/Koppeltaal/Mailbox')
    assert response is not None


def test_send_careplan_success_from_fixture_versioned_focal_resource(
        connector, transport, careplan_from_fixture):
    transport.expect(
        'POST',
        '/FHIR/Koppeltaal/Mailbox',
        respond_with='fixtures/bundle_post_careplan_ok.json')
    response_data = connector.send(
        'CreateOrUpdateCarePlan',
        careplan_from_fixture,
        careplan_from_fixture.patient)
    assert len(response_data) == 3
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource, response_data[0])
    assert response_data[0].fhir_link == (
        'https://example.com/fhir/Koppeltaal/CarePlan/1/'
        '_history/1970-01-01T01:01:01:01.1')
    assert response_data[1].fhir_link == (
        'https://example.com/fhir/Koppeltaal/Practitioner/1/'
        '_history/1970-01-01T01:01:01:01.1')
    assert response_data[2].fhir_link == (
        'https://example.com/fhir/Koppeltaal/Patient/1/'
        '_history/1970-01-01T01:01:01:01.1')

    response = transport.called.get('/FHIR/Koppeltaal/Mailbox')
    assert response is not None


def test_send_careplan_fail_message_response_error_fixture(
        connector, transport, careplan_from_fixture):
    transport.expect(
        'POST',
        '/FHIR/Koppeltaal/Mailbox',
        respond_with='fixtures/bundle_post_careplan_failed.json')
    with pytest.raises(koppeltaal.interfaces.MessageResponseError) as cm:
        connector.send(
            'CreateOrUpdateCarePlan',
            careplan_from_fixture,
            careplan_from_fixture.patient)

    response = transport.called.get('/FHIR/Koppeltaal/Mailbox')
    assert response is not None
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.MessageHeader,
        cm.value.message)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.MessageHeaderResponse,
        cm.value.message.response)


def test_send_careplan_operation_outcome_error_from_fixture(
        connector, transport, careplan_from_fixture):
    transport.expect(
        'POST',
        '/FHIR/Koppeltaal/Mailbox',
        respond_error='fixtures/operation_outcome.json')
    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as cm:
        connector.send(
            'CreateOrUpdateCarePlan',
            careplan_from_fixture,
            careplan_from_fixture.patient)

    response = transport.called.get('/FHIR/Koppeltaal/Mailbox')
    assert response is not None
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.OperationOutcome,
        cm.value.outcome)
    assert len(cm.value.outcome.issue) == 2
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Issue,
        cm.value.outcome.issue[0])
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource,
        cm.value.outcome.issue[0].resource)
    assert cm.value.outcome.issue[0].resource.fhir_link == \
        'https://example.com/fhir/Koppeltaal/Patient/1'
    assert cm.value.outcome.issue[0].details == (
        'Message Version mismatch: Please retrieve the latest version of '
        'this Message and apply changes before resubmit.')
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Issue,
        cm.value.outcome.issue[1])
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource,
        cm.value.outcome.issue[1].resource)
    assert cm.value.outcome.issue[1].resource.fhir_link == \
        'https://example.com/fhir/Koppeltaal/Practitioner/1'
    assert cm.value.outcome.issue[1].details == (
        'Message Version mismatch: Please retrieve the latest version of '
        'this Message and apply changes before resubmit.')


def test_updates_implicit_success_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 1
    for update in updates:
        with update:
            assert zope.interface.verify.verifyObject(
                koppeltaal.interfaces.IUpdate, update)
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.MessageHeader, update.message)
            assert update.message.event == 'CreateOrUpdateCarePlan'
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.CarePlan, update.data)
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.Patient, update.patient)

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))


def test_updates_explicit_success_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 1
    for update in updates:
        with update:
            assert zope.interface.verify.verifyObject(
                koppeltaal.interfaces.IUpdate, update)
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.MessageHeader, update.message)
            assert update.message.event == 'CreateOrUpdateCarePlan'
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.CarePlan, update.data)
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.Patient, update.patient)
            update.success()

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))


def test_updates_explicit_fail_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 1
    for update in updates:
        with update:
            assert zope.interface.verify.verifyObject(
                koppeltaal.interfaces.IUpdate, update)
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.MessageHeader, update.message)
            assert update.message.event == 'CreateOrUpdateCarePlan'
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.CarePlan, update.data)
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.Patient, update.patient)
            update.fail("I failed testing it.")

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            hamcrest.all_of(
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusStatus',
                    hamcrest.has_entry('valueCode', 'Failed')),
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusException',
                    hamcrest.has_entry('valueString', "I failed testing it."))
            )))


def test_updates_implicit_success_exception_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 1
    with pytest.raises(ValueError):
        for update in updates:
            with update:
                assert zope.interface.verify.verifyObject(
                    koppeltaal.interfaces.IUpdate, update)
                assert zope.interface.verify.verifyObject(
                    koppeltaal.definitions.MessageHeader, update.message)
                assert update.message.event == 'CreateOrUpdateCarePlan'
                assert zope.interface.verify.verifyObject(
                    koppeltaal.definitions.CarePlan, update.data)
                assert zope.interface.verify.verifyObject(
                    koppeltaal.definitions.Patient, update.patient)
                raise ValueError("I cannot write code.")

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'New'))
            ))


def test_updates_explicit_success_exception_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 1
    with pytest.raises(ValueError):
        for update in updates:
            with update:
                assert zope.interface.verify.verifyObject(
                    koppeltaal.interfaces.IUpdate, update)
                assert zope.interface.verify.verifyObject(
                    koppeltaal.definitions.MessageHeader, update.message)
                assert update.message.event == 'CreateOrUpdateCarePlan'
                assert zope.interface.verify.verifyObject(
                    koppeltaal.definitions.CarePlan, update.data)
                assert zope.interface.verify.verifyObject(
                    koppeltaal.definitions.Patient, update.patient)
                update.success()
                raise ValueError("I cannot write code.")

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            hamcrest.all_of(
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusStatus',
                    hamcrest.has_entry('valueCode', 'New')),
                hamcrest.not_(
                    koppeltaal.testing.has_extension(
                        '#ProcessingStatusException'))
            )))


def test_updates_error_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_error.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 0

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None

    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            hamcrest.all_of(
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusStatus',
                    hamcrest.has_entry('valueCode', 'Failed')),
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusException',
                    hamcrest.has_entry(
                        'valueString',
                        hamcrest.ends_with(
                            "RequiredMissing: 'startDate' "
                            "required but missing.")))
            )))


def test_updates_expected_event(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates(
        expected_events=['CreateOrUpdateCarePlan']))
    assert len(updates) == 1
    for update in updates:
        with update:
            assert update.message.event == 'CreateOrUpdateCarePlan'

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))


def test_updates_unexpected_event(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    # This is the response from the KT server after sending the fail.
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    # We ask for updates but only allow CreateOrUpdatePatient events. Since
    # the message we'll retrieve is a CreateOrUpdateCarePlan, we should see
    # that the message is acknowledged as "fail".
    updates = list(connector.updates(
        expected_events=['CreateOrUpdatePatient']))
    assert len(updates) == 0

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            hamcrest.all_of(
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusStatus',
                    hamcrest.has_entry('valueCode', 'Failed')),
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusException',
                    hamcrest.has_entry('valueString', "Event not expected"))
            )))


def test_updates_for_patient(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim&'
        'Patient=https%3A%2F%2Fapp.minddistrict.com%2Ffhir%2FKoppeltaal'
        '%2FPatient%2F1394433515',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim&'
        'Patient=https%3A%2F%2Fapp.minddistrict.com%2Ffhir%2FKoppeltaal'
        '%2FPatient%2F1394433515',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates(patient=(
        'https://app.minddistrict.com/fhir/Koppeltaal/Patient/1394433515')))
    assert len(updates) == 1
    for update in updates:
        with update:
            assert update.message.event == 'CreateOrUpdateCarePlan'

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))


def test_updates_for_event(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim&event=CarePlan',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim&event=CarePlan',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates(event='CarePlan'))
    assert len(updates) == 1
    for update in updates:
        with update:
            assert update.message.event == 'CreateOrUpdateCarePlan'

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))


def test_updates_unexpected_event_no_events_at_all(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    # This is the response from the KT server after sending the fail.
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    # We ask for updates but do not allow any events. Since the message we'll
    # retrieve is a CreateOrUpdateCarePlan, we should see that the message is
    # acknowledged as "fail".
    updates = list(connector.updates(expected_events=[]))
    assert len(updates) == 0

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            hamcrest.all_of(
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusStatus',
                    hamcrest.has_entry('valueCode', 'Failed')),
                koppeltaal.testing.has_extension(
                    '#ProcessingStatusException',
                    hamcrest.has_entry('valueString', "Event not expected"))
            )))


def test_updates_same_source_acked(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message_same_source.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    # The one incoming messages is from our "own" endpoint. This is not a
    # message we will process, but it is "acked" as success.

    updates = list(connector.updates())
    assert len(updates) == 0

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))


def test_software_name_and_version(
        connector, transport, careplan_from_fixture):
    transport.expect(
        'POST',
        '/FHIR/Koppeltaal/Mailbox',
        respond_with='fixtures/bundle_post_careplan_ok.json')
    connector.send(
        'CreateOrUpdateCarePlan',
        careplan_from_fixture,
        careplan_from_fixture.patient)
    sent = transport.called.get('/FHIR/Koppeltaal/Mailbox')
    assert sent is not None
    hamcrest.assert_that(
         sent,
         hamcrest.has_entry(
             'entry', hamcrest.has_item(
                 hamcrest.has_entry(
                     'content', hamcrest.has_entry(
                        'source', hamcrest.has_entries(
                            'endpoint', 'https://example.com/fhir/Koppeltaal',
                            'name', 'Koppeltaal Python Adapter Tests',
                            'software', (
                                'Koppeltaal Python Adapter; '
                                'Koppeltaal Python Adapter Tests Runner'),
                            'version', hamcrest.all_of(
                                hamcrest.starts_with('1.3.5.'),
                                hamcrest.contains_string('; 1.3.5.'))))))))



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

from __future__ import absolute_import

import koppeltaal.logger  # NOQA



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

from koppeltaal.fhir import resource
from koppeltaal import (interfaces, utils)


MARKER = object()


class Entry(resource.Entry):
    _atom_id = MARKER

    def __init__(self, packer, entry=None, model=None):
        if entry is not None:
            self._fhir_link = utils.json2links(entry).get('self')
            self._atom_id = entry.get('id')

            super(Entry, self).__init__(packer, content=entry['content'])
        if model is not None:
            super(Entry, self).__init__(packer, model=model)

    @property
    def atom_id(self):
        if self._atom_id is not MARKER:
            return self._atom_id

        link = self.fhir_link
        if link is None:
            link = self._packer.fhir_link(self._model, self.resource_type)
        self._atom_id = utils.strip_history_from_link(link)
        return self._atom_id

    def pack(self):
        entry = {
            "content": super(Entry, self).pack(),
            "id": self.atom_id}
        if self.fhir_link is not None:
            entry["link"] = [{"rel": "self", "href": self.fhir_link}]
        return entry

    def __eq__(self, other):
        if isinstance(other, dict) and 'reference' in other:
            return other['reference'] in filter(
                None,
                (self.history_less_fhir_link, self.fhir_link, self.atom_id))
        return super(Entry, self).__eq__(other)


class Bundle(resource.Resource):
    _create_entry = Entry

    def add_payload(self, response):
        if response['resourceType'] != 'Bundle' or 'entry' not in response:
            raise interfaces.InvalidBundle(self, response)
        for entry in response['entry']:
            self.items.append(self._create_entry(self.packer, entry=entry))

    def get_payload(self):
        assert self.domain is not None, 'Domain is required to create payloads'
        entries = list(self.pack())
        return {
            "resourceType": "Bundle",
            "id": "urn:uuid:{}".format(utils.uniqueid()),
            "updated": utils.now().isoformat(),
            "category": [{
                "term": "{}Domain#{}".format(
                    interfaces.NAMESPACE, self.domain),
                "label": self.domain,
                "scheme": "http://hl7.org/fhir/tag/security"
            }, {
                "term": "http://hl7.org/fhir/tag/message",
                "scheme": "http://hl7.org/fhir/tag"
            }],
            "entry": entries}

    def errors(self):
        errors = []
        for model in self.unpack():
            if interfaces.IBrokenFHIRResource.providedBy(model):
                errors.append(model)
        return errors



import sys

from setuptools import setup, find_packages


version = '1.3.5.12'


with open('README.md') as file:
    long_description = file.read()


install_requires = [
    'six',
    'lxml',
    'python-dateutil',
    'requests',
    'setuptools',
    'zope.interface >= 4.4',
    ]


tests_require = [
    'PyHamcrest >= 1.9',
    'selenium >= 3.8',
    ]


if sys.version_info.major == 2:
    install_requires.append('configparser')
    tests_require.append('mock==1.0.1')


setup(
    name='koppeltaal',
    version=version,
    license='AGPL',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    zip_safe=False,
    long_description=long_description,
    author='Koppeltaal community',
    url='https://github.com/Koppeltaal/Koppeltaal-Python-Connector',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: Dutch',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    install_requires=install_requires,
    extras_require={'test': tests_require},
    entry_points={
        'console_scripts': [
            'koppeltaal = koppeltaal.console:console'
            ],
        }
    )



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import zope.interface

from koppeltaal import (definitions, interfaces)


@zope.interface.implementer(interfaces.IFHIRResource)
class FHIRResource(object):
    fhir_link = None


@zope.interface.implementer(interfaces.IReferredFHIRResource)
class ReferredResource(FHIRResource):
    display = None

    def __init__(self, fhir_link=None, display=None):
        self.fhir_link = fhir_link
        self.display = display


@zope.interface.implementer(definitions.SubActivityDefinition)
class SubActivityDefinition(object):

    def __init__(
            self,
            active=True,
            description=None,
            identifier=None,
            name=None):
        self.active = active
        self.description = description
        self.identifier = identifier
        self.name = name


@zope.interface.implementer(definitions.ActivityDefinition)
class ActivityDefinition(FHIRResource):

    def __init__(
            self,
            description=None,
            identifier=None,
            application=None,
            is_active=True,
            is_archived=False,
            is_domain_specific=False,
            kind=None,
            launch_type='Web',
            name=None,
            performer=None,
            subactivities=None):
        self.description = description
        self.identifier = identifier
        self.application = application
        self.is_active = is_active
        self.is_archived = is_archived
        self.is_domain_specific = is_domain_specific
        self.kind = kind
        self.launch_type = launch_type
        self.name = name
        self.performer = performer
        self.subactivities = subactivities


@zope.interface.implementer(definitions.Name)
class Name(object):

    def __init__(
            self,
            family=None,
            given=None,
            prefix=None,
            suffix=None,
            text=None,
            use="official"):
        self.family = family
        self.given = given
        self.prefix = prefix
        self.suffix = suffix
        self.text = text
        self.use = use


@zope.interface.implementer(definitions.Contact)
class Contact(object):

    def __init__(
            self,
            system=None,
            value=None,
            use=None):
        self.system = system
        self.value = value
        self.use = use


@zope.interface.implementer(definitions.Identifier)
class Identifier(object):

    def __init__(
            self,
            system=None,
            value=None,
            use=None):
        self.system = system
        self.value = value
        self.use = use


@zope.interface.implementer(definitions.Participant)
@zope.interface.implementer(definitions.ActivityParticipant)
class Participant(object):

    def __init__(
            self,
            member=None,
            role=None,
            careteam=None):
        self.member = member
        self.role = role
        self.careteam = careteam


@zope.interface.implementer(definitions.OrganizationContactPerson)
class OrganizationContactPerson(object):

    def __init__(
            self,
            contacts=None,
            gender=None,
            name=None,
            purpose=None):
        self.contacts = contacts
        self.gender = gender
        self.name = name
        self.purpose = purpose


@zope.interface.implementer(definitions.Organization)
class Organization(FHIRResource):

    def __init__(
            self,
            active=None,
            address=None,
            category=None,
            contacts=None,
            contact_persons=None,
            identifiers=None,
            name=None,
            part_of=None):
        self.active = active
        self.address = address
        self.category = category
        self.contacts = contacts
        self.contact_persons = contact_persons
        self.identifiers = identifiers
        self.name = name
        self.part_of = part_of


@zope.interface.implementer(definitions.Patient)
class Patient(FHIRResource):

    def __init__(
            self,
            active=None,
            address=None,
            age=None,
            birth_date=None,
            care_providers=None,
            contacts=None,
            identifiers=None,
            gender=None,
            name=None,
            managing_organization=None):
        self.active = active
        self.address = address
        self.age = age
        self.birth_date = birth_date
        self.care_providers = care_providers
        self.contacts = contacts
        self.identifiers = identifiers
        self.gender = gender
        self.name = name
        self.managing_organization = managing_organization


@zope.interface.implementer(definitions.Practitioner)
class Practitioner(FHIRResource):

    def __init__(
            self,
            birth_date=None,
            contacts=None,
            identifiers=None,
            name=None,
            gender=None,
            organization=None):
        self.birth_date = birth_date
        self.contacts = contacts
        self.identifiers = identifiers
        self.name = name
        self.gender = gender
        self.organization = organization


@zope.interface.implementer(definitions.RelatedPerson)
class RelatedPerson(FHIRResource):

    def __init__(
            self,
            identifiers=None,
            patient=None,
            relationship=None,
            name=None,
            contacts=None,
            gender=None,
            address=None,
            photo=None):
        self.identifiers = identifiers
        self.patient = patient
        self.relationship = relationship
        self.name = name
        self.contacts = contacts
        self.gender = gender
        self.address = address
        self.photo = photo


@zope.interface.implementer(definitions.Goal)
class Goal(object):

    def __init__(
            self,
            description=None,
            notes=None,
            status=None):
        self.description = description
        self.status = status
        self.notes = notes


@zope.interface.implementer(definitions.CarePlanSubActivityStatus)
@zope.interface.implementer(definitions.SubActivity)
class SubActivity(object):

    def __init__(
            self,
            definition=None,
            status=None):
        self.definition = definition
        self.status = status


@zope.interface.implementer(definitions.ActivityDetails)
class ActivityDetails(object):

    def __init__(
            self,
            category=None,
            performers=None):
        self.category = category
        self.performers = performers


@zope.interface.implementer(definitions.Activity)
class Activity(object):

    def __init__(
            self,
            cancelled=None,
            definition=None,
            description=None,
            ends=None,
            finished=None,
            identifier=None,
            kind=None,
            notes=None,
            participants=None,
            planned=None,
            started=None,
            status=None,
            subactivities=None,
            details=None,
            prohibited=False):
        self.cancelled = cancelled
        self.definition = definition
        self.description = description
        self.finished = finished
        self.ends = ends
        self.identifier = identifier
        self.kind = kind
        self.notes = notes
        self.participants = participants
        self.planned = planned
        self.started = started
        self.status = status
        self.subactivities = subactivities
        self.details = details
        self.prohibited = prohibited


@zope.interface.implementer(definitions.CarePlan)
class CarePlan(FHIRResource):

    def __init__(
            self,
            activities=None,
            goals=None,
            participants=None,
            patient=None,
            status=None):
        self.activities = activities
        self.goals = goals
        self.participants = participants
        self.patient = patient
        self.status = status


@zope.interface.implementer(definitions.ProcessingStatus)
class ProcessingStatus(object):

    def __init__(
            self,
            exception=None,
            last_changed=None,
            status=None):
        self.exception = exception
        self.last_changed = last_changed
        self.status = status


@zope.interface.implementer(definitions.CarePlanActivityStatus)
class ActivityStatus(FHIRResource):

    def __init__(
            self,
            identifier=None,
            status=None,
            subactivities=None,
            percentage=None):
        self.identifier = identifier
        self.status = status
        self.subactivities = subactivities
        self.percentage = percentage


@zope.interface.implementer(definitions.MessageHeaderResponse)
class MessageHeaderResponse(object):

    def __init__(self,
                 identifier=None,
                 code=None):
        self.identifier = identifier
        self.code = code


@zope.interface.implementer(definitions.MessageHeaderSource)
class MessageHeaderSource(object):

    def __init__(
            self,
            endpoint=None,
            name=None,
            software=None,
            version=None):
        self.endpoint = endpoint
        self.name = name
        self.software = software
        self.version = version


@zope.interface.implementer(definitions.MessageHeader)
class MessageHeader(FHIRResource):

    def __init__(
            self,
            data=None,
            event=None,
            identifier=None,
            patient=None,
            response=None,
            source=None,
            status=None,
            timestamp=None):
        self.data = data
        self.event = event
        self.identifier = identifier
        self.patient = patient
        self.response = response
        self.source = source
        self.status = status
        self.timestamp = timestamp


@zope.interface.implementer(definitions.Period)
class Period(object):

    def __init__(
            self,
            start=None,
            end=None):
        self.start = start
        self.end = end


@zope.interface.implementer(definitions.CareTeam)
class CareTeam(FHIRResource):

    def __init__(
            self,
            identifier=None,
            status=None,
            name=None,
            subject=None,
            period=None,
            managing_organization=None):
        self.identifier = identifier
        self.status = status
        self.name = name
        self.subject = subject
        self.period = period
        self.managing_organization = managing_organization


@zope.interface.implementer(definitions.Address)
class Address(object):

    def __init__(
            self,
            city=None,
            country=None,
            line=None,
            period=None,
            state=None,
            text=None,
            use=None,
            zip=None):
        self.city = city
        self.country = country
        self.line = line
        self.period = period
        self.state = state
        self.text = text
        self.use = use
        self.zip = zip


@zope.interface.implementer(definitions.Issue)
class Issue(object):

    def __init__(
            self,
            severity=None,
            type=None,
            resource=None,
            details=None,
            location=None):
        self.severity = severity
        self.type = type
        self.resource = resource
        self.details = details
        self.location = location


@zope.interface.implementer(definitions.OperationOutcome)
class OperationOutcome(FHIRResource):

    def __init__(
            self,
            issue=None):
        self.issue = issue



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import collections
import datetime
import os.path
import six
import uuid

from configparser import ConfigParser
from six.moves.urllib.parse import urlparse


unicode = six.text_type


class UTC(datetime.tzinfo):

    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)


utc = UTC()


def strip_history_from_link(link):
    return link.split('/_history/', 1)[0]


def json2links(data):
    links = {}
    for link in data.get('link', []):
        links[link['rel']] = link['href']
    return links


def uniqueid():
    """Generate a unique ID.
    """
    return unicode(uuid.uuid4())


def messageid():
    """Generate a unique ID for a new message.
    """
    return uniqueid()


def now():
    return datetime.datetime.utcnow().replace(microsecond=0, tzinfo=utc)


Credentials = collections.namedtuple(
    'Credentials',
    ['url', 'username', 'password', 'domain', 'options'])


def get_credentials_from_file(name):
    # They're not passed in, so now look at ~/.koppeltaal.cfg.
    config = unicode(os.path.expanduser('~/.koppeltaal.cfg'))
    if not os.path.isfile(config):
        raise ValueError("Can't find ~/.koppeltaal.cfg")
    parser = ConfigParser()
    parser.read(config)
    if not parser.has_section(name):
        raise ValueError('No user credentials found in ~/.koppeltaal.cfg')
    url = parser.get(name, 'url')
    parsed = urlparse(url)
    if parsed.scheme != 'https' or \
            parsed.netloc == '' or \
            parsed.path != '' or \
            parsed.params != '' or \
            parsed.query != '' or \
            parsed.fragment != '':
        raise ValueError('Incorrect server URL.')
    username = parser.get(name, 'username')
    password = parser.get(name, 'password')
    domain = parser.get(name, 'domain')
    if not username or not password or not domain:
        raise ValueError('No user credentials found in ~/.koppeltaal.cfg')
    return Credentials(
        url, username, password, domain,
        dict(parser.items(name)))



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

from koppeltaal import (interfaces, logger)


def extensions(extensions, namespace=interfaces.NAMESPACE):
    """Update extensions from an older version of Koppeltaal to the
    currently supported one.
    """
    subactivities = extensions.get(namespace + 'CarePlan#SubActivity', [])
    for subactivity in subactivities:
        if 'valueString' in subactivity:
            logger.warning('Detected 1.0 subactivity in care plan.')
            subactivity['extension'] = [{
                'url': namespace + 'CarePlan#SubActivityIdentifier',
                'valueString': subactivity['valueString']}
            ]
            del subactivity['valueString']
